/*
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>
#include "SC_PlugIn.h"

// twiddle-ables
#define HEXAGON_SZ 7     // height of triangular mesh
//#define FEEDFORWARD
#define SELF_LOOP        // for control over tension
#define REFLECTING_EDGES // extra self-loops around edge, which invert signal
#define DAMPEN 0.9999    // loss
#define AUDIO_INPUT
#define TRIGGER_DURATION 1024 /* number of samples worth of white noise 
				 to inject */

// some constants from Brook Eaton's roto-drum
// http://www-ccrma.stanford.edu/~be/drum/drum.htm

#define DELTA 6.0 // distance between junctions
#define GAMMA 8.0 // wave speed

// A unit delay
typedef struct {
  float a;
  float b;
  int invert; // whether signal should be inverted
} t_delay;

// a scatter junction with up to 6 input and output delays, plus a
// self loop
typedef struct {
  int ins, outs;
  t_delay *in[6];
  t_delay *out[6];
#ifdef SELF_LOOP
  t_delay *self_loop;
#endif
} t_junction;


typedef struct {
  int x;
  int y;
} t_point;

// representation of a hexagon made up of triangles

/* hex - a 2d array of integers. The first dimension is the junction
 *   id, the second dimension is list of junction ids that junction
 *   connects to.  Only contains connections in one direction.  
 * twod - arranges the points in space, useful for visualising
 */

typedef struct {
  int **hex;
  t_point *twod;
  int junctions; // number of junctions
  int edges;     // number of connections
  int size;      // height of hexagon
} t_hexagon;


// supercollider stuff starts here...

// InterfaceTable contains pointers to functions in the host (server).
static InterfaceTable *ft;

// declare struct to hold unit generator state
struct Membrane : public Unit
{
  float yj; // junction admittence, calculated from tension parameter
#ifndef AUDIO_INPUT
  int triggered; // flag
  int excite;    // number of samples left in a triggered excitation
#endif
  t_hexagon *hexagon;
  t_junction *junctions;
  t_delay *delays;
  int delay_n;   // number of delays in mesh including self loops etc
};

// declare unit generator functions 
extern "C"
{
  void load(InterfaceTable *inTable);
  void Membrane_next_a(Membrane *unit, int inNumSamples);
  void Membrane_Ctor(Membrane* unit);
};

////////////////////////////////////////////////////////////////////

// push an integer on to the end of an array
void push(int *array, int value) {
  while(*array != 0) {
    ++array;
  }
  *array = value;
}

////////////////////////////////////////////////////////////////////

// construct a triangular mesh in the shape of a hexagon
// I worked this out on paper in an adhoc manner...  I wonder if 
// there is a nice library for this kind of thing?
t_hexagon *hexagon (int size) {
  int lengths[size];
  int x, y, i;
  int **hex;
  t_point *twod;
  int points = 0;
  int point = 0;
  int edges = 0;
  t_hexagon *result = (t_hexagon *) calloc(1, sizeof(t_hexagon));

  for (y = 0; y < size; ++y) {
    if (y <= (int) ((size - 1) / 2)) {
      lengths[y] = y + ((size - 1) / 2) + 1;
    }
    else {
      lengths[y] = size - y + ((size - 1) / 2);
    }
    points += lengths[y];
  }
  
  hex = (int **) calloc(points + 1, sizeof(int *));

  for (i = 0; i < points; i++) {
    hex[i] = (int *) calloc(1, sizeof(int) * 3);
  }

  twod = (t_point *) calloc(points, sizeof(t_point));
  
  for (y = 0; y < size; ++y) {
    int len = lengths[y];
    int offset = abs(((size-1)/2) - y);
    
    for (x = 0; x < len; ++x) {
      twod[point].y = y;
      twod[point].x = (x * 2) + offset;
      
      if (x < (len - 1)) {
	// 0 deg
	push(hex[point], point+1);
	edges++;
      }
      if (y < (size - 1)) {
	if (y < ((size - 1) / 2)) {
	  // 60 deg
	  push(hex[point], point + len + 1);
	  edges++;
	  // 120 deg
	  push(hex[point], point + len);
	  edges++;
	}
	else {
	  // 60 deg
	  if (x < (len - 1)) {
	    push(hex[point], point + len);
	    edges++;
	  }
	  // 120 deg
	  if (x > 0) {
	    push(hex[point], point + len - 1);
	    edges++;
	  }
	}
      }
      ++point;
    }
  }

#ifdef DEBUG
  for (i = 0; i < points; ++i) {
    printf("%d: ", i);
    for (x = 0; x < 3; ++x) {
      if (hex[i][x] == 0) {
	break;
      }
      printf(" %d ", hex[i][x]);
    }
    printf("\n");
  }
#endif

  result->hex = hex;
  result->edges = edges;
  result->junctions = points;
  result->size = size;
  result->twod = twod;

  return(result);
}

////////////////////////////////////////////////////////////////////

// execute one sample cycle over the mesh
void cycle(Membrane *unit, float input) {
  t_delay *delays = unit->delays;
  t_junction *junctions = unit->junctions;

  int i;

  // update all the junctions and waveguides
  for (i = 0; i < unit->hexagon->junctions; ++i) {
    t_junction *junction = &junctions[i];
    int j;
    double total = 0;

    double yc = unit->yj - junction->ins;

    for (j = 0; j < junction->ins; ++j) {
      total += junction->in[j]->b;
    }
    
#ifdef SELF_LOOP
    total = 2.0 * (total + (yc * junction->self_loop->b)) / unit->yj;
#else
    total *= (2.0 / ((double) junction->ins));
#endif

    total += input;

#ifdef DAMPEN
    total *= DAMPEN;
#endif

    for (j = 0; j < junction->outs; ++j) {
      junction->out[j]->a = total - junction->in[j]->b;
    }
#ifdef SELF_LOOP
    junction->self_loop->a = total - junction->self_loop->b;
#endif

  }

  // circulate the unit delays
  for (i = 0; i < unit->delay_n; ++i) {
    t_delay *delay = &delays[i];
    if (delay->invert) {
      delay->b = 0.0 - delay->a;
    }
    else {
      delay->b = delay->a;
    }
  }
}

////////////////////////////////////////////////////////////////////

// unit generator initialisation
void Membrane_Ctor(Membrane* unit)
{
  int **hex;
  int d = 0;
  int i = 0;
  int j = 0;
  t_delay *delay;

  // 1. Set calculation function -- always run at audio rate
  SETCALC(Membrane_next_a);
  
  // 2. Initialise state
#ifndef AUDIO_INPUT
  unit->triggered = 0;
  unit->excite = 0;
#endif
  unit->yj = 0;
  unit->hexagon = hexagon(HEXAGON_SZ);
  hex = unit->hexagon->hex;

  unit->delay_n = (unit->hexagon->edges * 2)
#ifdef REFLECTING_EDGES
    + (unit->hexagon->size * 6)
#endif
#ifdef SELF_LOOP
    + (unit->hexagon->junctions)
#endif
    ;

  // make all the junctions and delays
  unit->delays =
    (t_delay *) calloc(unit->delay_n, sizeof(t_delay));
  unit->junctions = 
    (t_junction *) calloc(unit->hexagon->junctions, 
			  sizeof(t_junction)
			  );

  for (i = 0; i < unit->hexagon->junctions; ++i) {
    t_junction *to;
    t_junction *from;
    int *n = hex[i];
    from = 
      (t_junction *) &unit->junctions[i];

    for (j = 0; j < 3; ++j) {
      if (*n == 0) {
	break;
      }
      to = 
	(t_junction *) &unit->junctions[*n];
      from->out[from->outs++] = &unit->delays[d];
      to->in[to->ins++] = &unit->delays[d];
      d++;
      to->out[to->outs++] = &unit->delays[d];
      from->in[from->ins++] = &unit->delays[d];
      d++;
      n++;
    }

#ifdef REFLECTING_EDGES		       
    while (from->ins < 6) {
      delay = &unit->delays[d++];
      delay->invert = 1;
      from->out[from->outs++] = delay;
      from->in[from->ins++] = delay;
    }
#endif
#ifdef SELF_LOOP
    delay = &unit->delays[d++];
    from->self_loop = delay;
#endif
  }

  // 3. Calculate one sample of output.
  // (why do this?)
  Membrane_next_a(unit, 1);
}


////////////////////////////////////////////////////////////////////

void Membrane_next_a(Membrane *unit, int inNumSamples) {
  // get the pointer to the output buffer
  float *out = OUT(0);
  int input_n = 0;
  // get the control rate input
#ifdef AUDIO_INPUT
  float *in = IN(input_n++);
#else
  float trigger = IN0(input_n++);
#endif
  float tension = IN0(input_n++);
  if (tension == 0) {
    // default tension
    tension =  0.0001;
  }
  unit->yj = 2 * DELTA * DELTA / (tension * tension * GAMMA * GAMMA);
  
#ifndef AUDIO_INPUT
  if (trigger >= 0.5 && (! unit->triggered)) {
    unit->triggered = 1;
    unit->excite = TRIGGER_DURATION;
    
  }
  else if (trigger < 0.5 && unit->triggered) {
    unit->triggered = 0;
  }
#endif

  for (int k=0; k < inNumSamples; ++k) {
    float input = 0.0;
#ifdef AUDIO_INPUT
    input = in[k] * 0.05;
#else
    if (unit->excite > 0) {
      input = (0.01 - (((double) rand() / RAND_MAX) * 0.02));
      unit->excite--;
    }
#endif
    cycle(unit, input);
    // send the first waveguide to the output
    out[k] = unit->delays[0].b + unit->delays[1].b;
  }
}

////////////////////////////////////////////////////////////////////

// the load function is called by the host when the plug-in is loaded
void load(InterfaceTable *inTable) {
  ft = inTable;
  DefineSimpleUnit(Membrane);
}

////////////////////////////////////////////////////////////////////
